#include "xmltk.h"

typedef struct
{
    int     iIndent;
    int     iIndentSize;
    bool    fLastWasClose;
    bool    fBracketOpen;
    bool    fAttribute;
    bool    fInCDATA;
    bool    fFirstElement;
} TSAXWriterState; 

class CTSAX2XML : public IXmlWriter
{
public:
    // *** IUnknownCL methods ***
    CL_STDMETHOD(QueryInterface) (THIS_ RCLIID riid, void **ppvObj);
    CL_STDMETHOD_(ULONG,AddRef) (THIS);
    CL_STDMETHOD_(ULONG,Release) (THIS);

    // ITSAXContentHandler methods.
    CL_STDMETHOD(startDocument) ();
    CL_STDMETHOD(endDocument) ();
    CL_STDMETHOD(startElement) (XTOKEN xtName);
    CL_STDMETHOD(endElement) (XTOKEN xtName);
    CL_STDMETHOD(attribute) (XTOKEN xtName, char *pszChars, int cchChars);
    CL_STDMETHOD(characters) (char *pszChars, int cchChars);
    CL_STDMETHOD(startCDATA) (void);
    CL_STDMETHOD(endCDATA) (void);
    
    CL_STDMETHOD_(void, SetBufferMode) (bool fWriteThrough);
    CL_STDMETHOD_(void, ClearBufferMode) (void);
    CL_STDMETHOD_(bool, FBufferMode) (void) { return m_fBuffer; }
    CL_STDMETHOD_(void, FlushBuffer) (void);
    CL_STDMETHOD_(size_t, GetSize) (void);

    // IXmlWriter methods.
    CL_STDMETHOD_(void, SetOutputStream) (IStream * pstOut)
    {
        m_pfOut = NULL;
        ATOMICRELEASE(m_pstOut);
        ASSIGN(m_pstOut, pstOut);
    }
    CL_STDMETHOD_(void, SetOutputFile) (FILE * pfOut)
    {
        ATOMICRELEASE(m_pstOut);
        m_pfOut = pfOut;
    }

    CTSAX2XML(int iIndentSize)
    {
        m_cRef = 1;

        m_pfOut  = NULL;
        m_pstOut = NULL;

        GetGlobalTokenTable(&IID_ITokenTable, (void**)&m_ptt);
        m_fBuffer = false;
        m_fWriteThrough = true;

        m_state.iIndent       = 0;
        m_state.iIndentSize   = iIndentSize;
        m_state.fLastWasClose = false;
        m_state.fBracketOpen  = false;
        m_state.fAttribute    = false;
        m_state.fInCDATA      = false;
        m_state.fFirstElement = true;
        SaveState();
    }

private:
    ULONG m_cRef;

    virtual ~CTSAX2XML()
    {
        ATOMICRELEASE(m_ptt);
        ATOMICRELEASE(m_pstOut);
    }

    void _PrintIndent()
    {
        int iIndent = m_state.iIndent * m_state.iIndentSize;
        for (int i = 0; i < iIndent; i++)
        {
            Write(" ");
        }
    }

    void _CloseBracketIfNeeded()
    {
        if (m_state.fBracketOpen)
        {
            Write(">");
            m_state.fBracketOpen = false;
            m_state.fAttribute = false;
        }
    }

    // Write some text out, either to the output stream or to the buffer.
    void Write(char * pszText, int cchText);
    void Write(char * pszText) { Write(pszText, strlen(pszText)); }
    
    void SaveState() { memcpy(&m_savedState, &m_state, sizeof(TSAXWriterState)); }
    void RestoreState() { memcpy(&m_state, &m_savedState, sizeof(TSAXWriterState)); }

    FILE *m_pfOut;
    IStream * m_pstOut;
    ITokenTable *m_ptt;

    TSAXWriterState m_state;
    TSAXWriterState m_savedState;

    bool m_fBuffer;
    bool m_fWriteThrough;
    CStringBuffer m_buffer;
};

// *** IUnknown methods ***
bool CTSAX2XML::QueryInterface(RCLIID riid, void **ppvObj)
{
    if (IsEqualCLIID(riid, &IID_IUnknownCL) ||
        IsEqualCLIID(riid, &IID_ITSAXContentHandler) ||
        IsEqualCLIID(riid, &IID_IBufferedTSAXContentHandler))
    {
        *ppvObj = static_cast<IBufferedTSAXContentHandler*>(this);
    }
    else 
    {
        *ppvObj = NULL;
        return false;
    }
    
    AddRef();
    return true;
}

ULONG CTSAX2XML::AddRef()
{
    return ++m_cRef;
}

ULONG CTSAX2XML::Release()
{
    if (--m_cRef > 0)
    {
        return m_cRef;
    }
    delete this;
    return 0;
}

void CTSAX2XML::Write(char * pszText, int cchText)
{
    if (m_fBuffer)
        m_buffer.Write(pszText, cchText);

    if (m_fWriteThrough)
    {
        if (m_pstOut)
        {
            assert(m_pfOut == NULL);
            m_pstOut->Write(pszText, cchText);
        }
        else
            fwrite(pszText, cchText, 1, m_pfOut);
    }
}

void CTSAX2XML::SetBufferMode(bool fWriteThrough)
{
    if (!m_fBuffer)
    {
        m_buffer.Clear();
        _CloseBracketIfNeeded();

        m_fBuffer = true;
        m_fWriteThrough = fWriteThrough;
        if (!m_fWriteThrough)
        {
            // We might clear the buffer afterwards, so we need to save the current state.
            SaveState();
        }
    }
}

void CTSAX2XML::ClearBufferMode()
{
    if (m_fBuffer && !m_fWriteThrough)
    {
        // We cleared the buffer contents, restore the state to the last good one.
        RestoreState();
        m_fWriteThrough = true;
    }
    m_fBuffer = false;
}

void CTSAX2XML::FlushBuffer()
{
    if (!m_fBuffer)
        _CloseBracketIfNeeded();

    if (m_pstOut)
    {
        char * pszString = const_cast<char *>(m_buffer.SzString());
        m_pstOut->Write(pszString, strlen(pszString));
    }
    else
        fputs(m_buffer.SzString(), m_pfOut);
    
    if (m_fBuffer && !m_fWriteThrough)
    {
        // We flushed something to the output, update the saved state.
        SaveState();
    }
}

size_t CTSAX2XML::GetSize()
{
    return (m_fBuffer ? m_buffer.GetSize() : 0);
}

bool FIsWhitespace(char * pszChars, int cchChars)
{
    for (int iChar = 0 ; iChar < cchChars ; iChar++, pszChars++)
    {
        if (*pszChars != '\n' || *pszChars != '\t' || *pszChars != ' ')
            return false;
    }

    return true;
}

bool CTSAX2XML::characters(char* pszChars, int cchChars)
{
    _CloseBracketIfNeeded();

    if (m_state.iIndentSize != 0 && !m_state.fInCDATA && m_state.fLastWasClose && 
        FIsWhitespace(pszChars, cchChars))
    {
        // Skip whitespace if we are doing the indentation ourselves.
        return true;
    }

    m_state.fLastWasClose = false;
    Write(pszChars, cchChars);

    return true;
}

bool CTSAX2XML::startCDATA(void)
{
    m_state.fInCDATA = true;

    _CloseBracketIfNeeded();
    Write("<![CDATA[");

    return true;
}

bool CTSAX2XML::endCDATA(void)
{
    m_state.fInCDATA = false;

    // Close off the CDATA section.
    Write("]]>");
    return true;
}

bool CTSAX2XML::endDocument()
{
    Write("\n"); 
    return true;
}

bool CTSAX2XML::startDocument()
{
    return true;
}

bool CTSAX2XML::endElement(XTOKEN xtName)
{
    m_state.iIndent--;

    if (m_state.fBracketOpen && !m_state.fAttribute)
    {
        //
        // the element tag is still open, and there were no
        // attributes, so we can output it in "<name/>" form.
        //
        m_state.fBracketOpen = false;
        Write("/>");
    }
    else
    {
        _CloseBracketIfNeeded();
        
        if (m_state.fLastWasClose && m_state.iIndentSize != 0)
        {
            Write("\n");
            _PrintIndent();
        }
        
        Write("</");
        Write(g_ptt->StrFromXTOKEN(xtName));
        Write(">");
    }

    m_state.fLastWasClose = true;
    
    return true;
}

bool CTSAX2XML::startElement(XTOKEN xtName)
{
    _CloseBracketIfNeeded();

    m_state.fLastWasClose = false;
    m_state.fBracketOpen  = true;

    if (!m_state.fFirstElement && m_state.iIndentSize != 0)
    {
        Write("\n");
        _PrintIndent();
    }
    m_state.iIndent++;
    m_state.fFirstElement = false; // This way, we will nicely put a newline between fragments.

    Write("<");
    Write(g_ptt->StrFromXTOKEN(xtName));
    return true;
}

bool CTSAX2XML::attribute(XTOKEN xtName, char *pszChars, int cchChars)
{
    m_state.fAttribute = true;
    
    Write(g_ptt->StrFromXTOKEN(xtName));
    Write("=\"");
    Write(pszChars, cchChars);
    Write("\"");
    return true;
}

IXmlWriter * CreateXmlWriter(int iIndent)
{
    return (new CTSAX2XML(iIndent));
}