// BUGS:
// * If an error occurs, the interfaces don't get cleaned up.
//   Solution: We should implement auto-pointers.

#include "xmltk.h"

Variable * g_pvRoot = NULL;

enum CmpOp
{
    CMP_EQ = 0,
    CMP_NE,
    CMP_LT,
    CMP_GT,
    CMP_LTE,
    CMP_GTE,
    CMP_LIKE,
    CMP_MAX     // This should be always be the last item.
};

typedef struct
{
    Variable * pvarElement;
    Variable * pvarKey;
    CmpOp cmpOp;
    union // Value of the key comparison.
    {
        long lCmpValue;
        char * pszCmpValue;
    };
} KeyInfo;

class CKeyInfoHolder
{
public:
    // CVariableHolder constructor and destructor.
    CKeyInfoHolder(KeyInfo * pKeyInfo) { m_pKeyInfo = pKeyInfo; m_pNext = NULL; }
    ~CKeyInfoHolder();

    KeyInfo * Find(Variable * pvarElement);
    void Add(CKeyInfoHolder * pNew);

private:
    KeyInfo         * m_pKeyInfo;
    CKeyInfoHolder  * m_pNext;
};

CKeyInfoHolder::~CKeyInfoHolder()
{
    delete m_pKeyInfo;
    if (m_pNext != NULL)
        delete m_pNext;
}

KeyInfo * CKeyInfoHolder::Find(Variable * pvarElement)
{
    for (CKeyInfoHolder * pHolder = this ; pHolder != NULL ; pHolder = pHolder->m_pNext)
    {
        if (pHolder->m_pKeyInfo->pvarElement == pvarElement)
            return pHolder->m_pKeyInfo;
    }

    return NULL;
}

void CKeyInfoHolder::Add(CKeyInfoHolder * pNew)
{
    assert(!pNew->m_pNext);

    for (CKeyInfoHolder * pHolder = this ; pHolder->m_pNext != NULL ; pHolder = pHolder->m_pNext);
    pHolder->m_pNext = pNew;
}

class CXGrepHandler : public CTSAXContentHandler
{
public:
    // XGrepHandler constructor and destructor.
    CXGrepHandler(IBufferedTSAXContentHandler * pchOut);
    virtual ~CXGrepHandler();

    CL_STDMETHOD(startContext) (Variable *pv);
    CL_STDMETHOD(endContext) (Variable *pv);
    CL_STDMETHOD(startElement) (XTOKEN xtName);
    CL_STDMETHOD(endElement) (XTOKEN xtName);
    CL_STDMETHOD(attribute) (XTOKEN xtName, char *pszChars, int cchChars);
    CL_STDMETHOD(characters) (char *pszChars, int cchChars);
    CL_STDMETHOD(startCDATA) (void);
    CL_STDMETHOD(endCDATA) (void);

    // Context variable.
    void SetContext(Variable * pvarContext) { m_pvarContext = pvarContext; }

    // Key information.
    CmpOp GetCmpOp(char * pszCmpOp);
    void SetKeyInfo(Variable * pvarElement, Variable * pvarKey, CmpOp cmpOp, char * pszCmpValue);
    void SetKeyInfo(Variable * pvarElement);
    
private:
    enum State
    {
        STATE_NO_CONTEXT = 0,
        STATE_START_CONTEXT,
        STATE_WRITE_CONTEXT,
        STATE_IN_CONTEXT,
        STATE_IN_ELEMENT,
        STATE_CHECK_KEY,
    };

    void SetBufferMode(bool fBufferMode)
    { 
        if (fBufferMode)
            m_pchOut->SetBufferMode(false);
        else
            m_pchOut->ClearBufferMode();
    }
    bool FBufferMode() { return m_pchOut->FBufferMode(); }
    void FlushBuffer() { m_pchOut->FlushBuffer(); }
    bool FCheckKey(const char * pszValue, int cchValue);

    State                         m_state;
    XTOKEN                        m_xtContext;    // This is the element that represents the context.
    Variable                    * m_pvarContext;  // Context variable.
    CKeyInfoHolder              * m_pKeys;        // All the information about element and key context.
    KeyInfo                     * m_pKeyInfoCur;  // Current element and key context information.
    IBufferedTSAXContentHandler * m_pchOut;       // Output stream handler, can be switched to buffer mode.
};

CXGrepHandler::CXGrepHandler(IBufferedTSAXContentHandler * pchOut)
{
    m_state       = STATE_NO_CONTEXT;
    m_pvarContext = NULL;
    m_pKeys       = NULL;
    m_pKeyInfoCur = NULL;
    
    ASSIGN(m_pchOut, pchOut);
}

CXGrepHandler::~CXGrepHandler()
{
    if (m_pKeys)
        delete m_pKeys;
    ATOMICRELEASE(m_pchOut); 
}

CmpOp CXGrepHandler::GetCmpOp(char * pszCmpOp)
{
    int cchCmpOp = strlen(pszCmpOp);
    if (cchCmpOp == 1)
    {
        switch (*pszCmpOp)
        {
            case '=':  return CMP_EQ;
            case '<':  return CMP_LT;
            case '>':  return CMP_GT;
            default: break;
        }
    }
    else if (cchCmpOp == 2)
    {
        switch (*pszCmpOp)
        {
        case '!':
            if (*++pszCmpOp != '=')
                break;
            return CMP_NE;
        case '<':
            if (*++pszCmpOp != '=')
                break;
            return CMP_LTE;
        case '>':
            if (*++pszCmpOp != '=')
                break;
            return CMP_GTE;
        default:
            break;
        }
    }
    else if (cchCmpOp == 4 && strncmp(pszCmpOp, "LIKE", 4))
    {
        return CMP_LIKE;
    }

    throw CXmlToolkitException("Invalid comparison operator.");
}

void CXGrepHandler::SetKeyInfo(Variable * pvarElement)
{
    assert(pvarElement != NULL);
    KeyInfo * pKeyInfo = new KeyInfo;
    pKeyInfo->pvarElement = pvarElement;
    pKeyInfo->pvarKey = NULL;

    CKeyInfoHolder * pHolder = new CKeyInfoHolder(pKeyInfo);
    if (m_pKeys == NULL)
        m_pKeys = pHolder;
    else
        m_pKeys->Add(pHolder);
}

void CXGrepHandler::SetKeyInfo(Variable * pvarElement, Variable * pvarKey, CmpOp cmpOp, char * pszCmpValue)
{
    assert(pvarElement != NULL && pvarKey != NULL && cmpOp < CMP_MAX && pszCmpValue != NULL);

    // We currently don't support LIKE.
    if (cmpOp == CMP_LIKE)
        throw new CXmlToolkitException("CXGrepHandler: 'LIKE' comparison not yet implemented.");
    
    KeyInfo * pKeyInfo = new KeyInfo;
    pKeyInfo->pvarElement = pvarElement;
    pKeyInfo->pvarKey = pvarKey;
    pKeyInfo->cmpOp = cmpOp;

    if (cmpOp != CMP_LIKE && cmpOp != CMP_EQ)
        pKeyInfo->lCmpValue = atol(pszCmpValue);
    else
        pKeyInfo->pszCmpValue = pszCmpValue;

    CKeyInfoHolder * pHolder = new CKeyInfoHolder(pKeyInfo);
    if (m_pKeys == NULL)
        m_pKeys = pHolder;
    else
        m_pKeys->Add(pHolder);
}

bool CXGrepHandler::startContext(Variable * pvar)
{
    if (pvar != g_pvRoot)
    {
        switch (m_state)
        {
        case STATE_NO_CONTEXT:
            // We didn't get any startContext yet, so this has to be the start of the 'xgrep context'.
            m_state = STATE_START_CONTEXT;
            break;
        case STATE_START_CONTEXT:
            // We don't allow setting an element filter that is identical to the context filter.
            throw new CXmlToolkitException("An element filter cannot be the same as the context filter.");
        case STATE_WRITE_CONTEXT:
        case STATE_IN_CONTEXT:
            // We matched a context filter already, so this has to be an element filter context.
            m_pKeyInfoCur = m_pKeys->Find(pvar);
            assert(m_pKeyInfoCur != NULL);
            SetBufferMode(true);
            m_state = STATE_IN_ELEMENT;
            break;
        case STATE_IN_ELEMENT:
            m_state = STATE_CHECK_KEY;
            break;
        case STATE_CHECK_KEY:
            // We should never get a startContext call in this state.
            assert(!"Invalid context call received!");
            break;
        }
    }

    return true;
}

bool CXGrepHandler::endContext(Variable * pvar)
{
    if (pvar == g_pvRoot)
        return true;

    switch (m_state)
    {
    case STATE_IN_CONTEXT:
        if (m_pKeyInfoCur != NULL && pvar == m_pKeyInfoCur->pvarElement)
        {
            m_pKeyInfoCur = NULL;
            break;
        }
        // Fall through...
    case STATE_WRITE_CONTEXT:
        // Close the context element and get out of here.
        m_pchOut->endElement(m_xtContext);
        m_state = STATE_NO_CONTEXT;
        break;
    case STATE_IN_ELEMENT:
        if (pvar != m_pKeyInfoCur->pvarKey)
        {
            // We are finished with this element, flush the buffer if we were buffering something.
            if (FBufferMode())
            {
                FlushBuffer();
                SetBufferMode(false);
            }

            m_state = STATE_IN_CONTEXT;
        }
        else if (!FBufferMode())
        {
            // If we are not buffering, then go back to context mode.
            m_state = STATE_IN_CONTEXT;
        }
        break;
    case STATE_NO_CONTEXT:
    case STATE_START_CONTEXT:
    case STATE_CHECK_KEY:
        assert(!"Invalid context call received!");
        break;
    }
        
    return true;
}

bool CXGrepHandler::startElement(XTOKEN xtName)
{
    switch (m_state)
    {
    case STATE_CHECK_KEY:
        // TODO: Implement concatenation of text. Right now we consider there is no match.
        SetBufferMode(false);
        m_state = STATE_IN_CONTEXT;
        break;
    case STATE_WRITE_CONTEXT:
        // This is the first start element call after the context started, we are now in the context.
        m_state = STATE_IN_CONTEXT;
        break;
    case STATE_START_CONTEXT:
        m_xtContext = xtName;
        m_state = STATE_WRITE_CONTEXT;
        // Pass through...
    case STATE_NO_CONTEXT:
    case STATE_IN_ELEMENT:
        m_pchOut->startElement(xtName);
    }

    return true;
}

bool CXGrepHandler::endElement(XTOKEN xtName)
{
    assert(m_state != STATE_CHECK_KEY);
    switch (m_state)
    {
    case STATE_NO_CONTEXT:
    case STATE_IN_ELEMENT:
        m_pchOut->endElement(xtName);
        break;
    case STATE_START_CONTEXT:
        assert(!"Invalid context call received!");
        break;
    }
    
    return true;
}

bool CXGrepHandler::attribute(XTOKEN xtName, char *pszChars, int cchChars)
{
    switch (m_state)
    {
    case STATE_CHECK_KEY:
        if (!FCheckKey(pszChars, cchChars))
            SetBufferMode(false);
        else
            m_pchOut->attribute(xtName, pszChars, cchChars);
        m_state = STATE_IN_ELEMENT;
        break;
    case STATE_NO_CONTEXT:
    case STATE_IN_ELEMENT:
    case STATE_WRITE_CONTEXT:
        m_pchOut->attribute(xtName, pszChars, cchChars);
        break;
    case STATE_START_CONTEXT:
        throw new CXmlToolkitException("The context specified must be an element filter.");
    }
  
    return true;
}

bool CXGrepHandler::characters(char *pszChars, int cchChars)
{
    switch (m_state)
    {
    case STATE_CHECK_KEY:
        if (!FCheckKey(pszChars, cchChars))
            SetBufferMode(false);
        else
            m_pchOut->characters(pszChars, cchChars);
        m_state = STATE_IN_ELEMENT;
        break;
    case STATE_WRITE_CONTEXT:
        // This is the first characters call after the context started, we are now in the context.
        m_state = STATE_IN_CONTEXT;
        break;
    case STATE_NO_CONTEXT:
    case STATE_IN_ELEMENT:
        m_pchOut->characters(pszChars, cchChars);
        break;
    case STATE_START_CONTEXT:
        throw new CXmlToolkitException("The context specified must be an element filter.");
    }

    return true;
}

bool CXGrepHandler::startCDATA(void)
{
    assert(m_state != STATE_START_CONTEXT && m_state != STATE_CHECK_KEY);
    switch (m_state)
    {
    case STATE_NO_CONTEXT:
    case STATE_IN_ELEMENT:
        m_pchOut->startCDATA();
        break;
    case STATE_WRITE_CONTEXT:
        // This is the first start cdata call after the context started, we are now in the context.
        m_state = STATE_IN_CONTEXT;
        break;
    case STATE_START_CONTEXT:
    case STATE_CHECK_KEY:
        assert(!"Invalid start CDATA call received!");
        break;
    }
    
    return true;
}

bool CXGrepHandler::endCDATA(void)
{
    switch (m_state)
    {
    case STATE_NO_CONTEXT:
    case STATE_IN_ELEMENT:
        m_pchOut->endCDATA();
        break;
    case STATE_START_CONTEXT:
    case STATE_WRITE_CONTEXT:
    case STATE_CHECK_KEY:
        assert(!"Invalid start CDATA call received!");
        break;
    }
    
    return true;
}

//-------------------------------------------------------------------------------------------------
// XGrepHandler private methods.
//-------------------------------------------------------------------------------------------------

bool CXGrepHandler::FCheckKey(const char * pszValue, int cchValue)
{
    assert(m_pKeyInfoCur->cmpOp != CMP_LIKE);

    long lValue = 0;
    if (m_pKeyInfoCur->cmpOp != CMP_EQ)
    {
        for (int iValuePos = 0 ; iValuePos < cchValue ; iValuePos++, pszValue++)
        {
            if (*pszValue < '0' || *pszValue > '9')
            {
                // This particular value is not a number. Consider this not matching.
                return false;
            }
            
            lValue = lValue*10 + (*pszValue - '0');
        }
    }

    switch (m_pKeyInfoCur->cmpOp)
    {
        case CMP_EQ: 
            return (strlen(m_pKeyInfoCur->pszCmpValue) == cchValue &&
                    strncmp(m_pKeyInfoCur->pszCmpValue, pszValue, cchValue) == 0);
        case CMP_NE: return (lValue != m_pKeyInfoCur->lCmpValue);
        case CMP_LT: return (lValue < m_pKeyInfoCur->lCmpValue);
        case CMP_GT: return (lValue > m_pKeyInfoCur->lCmpValue);
        case CMP_GTE: return (lValue >= m_pKeyInfoCur->lCmpValue);
        case CMP_LTE: return (lValue <= m_pKeyInfoCur->lCmpValue);
        default: assert(!"Unrecogmized comparison operator!");
    }

    return false;
}

typedef struct
{
    CXGrepHandler * pHandler;
    IDfaFilter * pdfa;
    char * pszXmlFile;
} XGrepArgs;

bool FProcessInput(int argc, char * argv[], XGrepArgs * pArgs)
{
    if (argc < 4)
        return false;

    Variable * pvarContext = NULL;
    Variable * pvarElement = NULL;
    for (int iArg = 1 ; iArg < argc ; iArg++)
    {
        char * pszArg = argv[iArg];
        if (*pszArg != '-')
        {
            pArgs->pszXmlFile = pszArg;

            if (pvarElement != NULL)
            {
                // This is an element xpath without key.
                pArgs->pHandler->SetKeyInfo(pvarElement);
            }

            return true;
        }
        pszArg++;
        
        switch (*pszArg)
        {
        case 'c':
            pvarContext = pArgs->pdfa->RegisterQuery(NULL, argv[++iArg], true);
            pArgs->pHandler->SetContext(pvarContext);
            break;
        case 'e':
            if (pvarContext == NULL)
                return false;

            if (pvarElement != NULL)
            {
                // This is an element xpath without key.
                pArgs->pHandler->SetKeyInfo(pvarElement);
            }

            pvarElement = pArgs->pdfa->RegisterQuery(pvarContext, argv[++iArg], true);
            break;
        case 'k':
        {
            if (pvarElement == NULL)
                return false;
            
            // Set the key information.
            char * pszKey = argv[++iArg];
            char * pszCmpOp = strchr(pszKey, ' ');
            if (pszCmpOp == NULL)
                return false;
            *pszCmpOp++ = '\0';
            char * pszCmpValue = strchr(pszCmpOp, ' ');
            if (pszCmpValue == NULL)
                return false;
            *pszCmpValue++ = '\0';
            pArgs->pHandler->SetKeyInfo(
                pvarElement, pArgs->pdfa->RegisterQuery(pvarElement, pszKey, true), 
                pArgs->pHandler->GetCmpOp(pszCmpOp), pszCmpValue);

            pvarElement = NULL;
            break;
        }
        default:
            return false;
        }
    }

    return false;
}

int main(int argc, char* argv[])
{
    InitGlobalTokenTable();
    try 
    {
        // Create my handler object that works on a stdout stream.
        IXmlWriter * pchOut = CreateXmlWriter();
        pchOut->SetOutputFile(stdout);
        CXGrepHandler handler(pchOut);

        // Create a new XPath processor, setting the handler on it.
        IDfaFilter * pdfa = CreateDfaFilter(&handler);

        XGrepArgs args;
        args.pdfa = pdfa;
        args.pHandler = &handler;
        if (FProcessInput(argc, argv, &args))
        {
            // Parse the xml file.
            CXmlToolkit xmltk(args.pszXmlFile, pdfa);
            pdfa->SetXmlParser(xmltk.GetXmlParser());

            g_pvRoot = pdfa->RegisterQuery(NULL, "/", true);

            xmltk.Parse();
        }
        else
        {
            cout << "Usage: xgrep -c context [-e <element to include> [-k \"<key comparison>\"]]+ <input file>\n";
            cout << "  <element to include> := <xpath to an element>\n";
            cout << "  <key comparison> := <xpath to key> <comparison operator> <numeric value>\n";
            cout << "  <comparison operator> := '=' | '!=' | '<' | '>' | '<=' | '>='\n";
        }
        
        // Cleanup.
        ATOMICRELEASE(pdfa);
        ATOMICRELEASE(pchOut);
    }
    catch (CXmlToolkitException &e)
    {
        printf("\n\n !! Error: %s !!", e.PszMessage());
    }
    CleanupGlobalTokenTable();

    return 0;
}